import numpy as np
import pytest

from bmi_tester.units import Units

# try:
#     from pymt import cfunits
# except (FileNotFoundError, ModuleNotFoundError):
#     cfunits = None


# @pytest.mark.dependency()
# @pytest.mark.dependency(depends=["test_get_var_location"])
# def test_get_var_grid(initialized_bmi, var_name):
#     """Test the grid of the variables."""
#     if initialized_bmi.get_var_location(var_name) == "none":
#         pytest.skip(f"var, {var_name}, is not located on a grid")
#
#     gid = initialized_bmi.get_var_grid(var_name)
#     assert isinstance(gid, int)


def test_get_var_itemsize(initialized_bmi, var_name):
    """Test getting a variable's itemsize"""
    itemsize = initialized_bmi.get_var_itemsize(var_name)
    assert itemsize > 0


# @pytest.mark.dependency()
def test_get_var_nbytes(initialized_bmi, var_name):
    """Test getting a variable's nbytes"""
    nbytes = initialized_bmi.get_var_nbytes(var_name)
    assert nbytes > 0


# @pytest.mark.dependency()
def test_get_var_location(initialized_bmi, var_name):
    """Test getting a variable's grid location"""
    location = initialized_bmi.get_var_location(var_name)
    assert isinstance(location, str)
    assert location in ("node", "edge", "face", "none")


# @pytest.mark.dependency(depends=["test_get_var_location"])
def test_var_on_grid(initialized_bmi, var_name):
    loc = initialized_bmi.get_var_location(var_name)
    if initialized_bmi.get_var_location(var_name) == "none":
        pytest.skip(f"var, {var_name}, is not located on a grid")

    gid = initialized_bmi.get_var_grid(var_name)
    if loc == "node":
        assert initialized_bmi.get_grid_node_count(gid) > 0
    elif loc == "edge":
        assert initialized_bmi.get_grid_edge_count(gid) > 0
    elif loc == "face":
        assert initialized_bmi.get_grid_face_count(gid) > 0


def test_get_var_type(initialized_bmi, var_name):
    """Test getting a variable's data type"""
    dtype = initialized_bmi.get_var_type(var_name)
    assert isinstance(dtype, str)

    try:
        np.empty(1, dtype=dtype)
    except TypeError:
        raise AssertionError(
            "get_var_type: bad data type name ({dtype})".format(dtype=dtype)
        )


# @pytest.mark.skipif(cfunits is None, reason="cfunits is broken on this platform")
def test_get_var_units(initialized_bmi, var_name):
    """Test the units of the variables."""
    units = initialized_bmi.get_var_units(var_name)
    assert isinstance(units, str)
    unit_system = Units()
    assert unit_system.is_valid(units)
    # cfunits.Units(units)



import importlib
import os

import pytest
from scripting import cp


# from .conftest import INPUT_FILE, Bmi

def load_component(entry_point):
    module_name, cls_name = entry_point.split(":")

    component = None
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        raise
    else:
        try:
            component = module.__dict__[cls_name]
        except KeyError:
            raise ImportError(cls_name)

    return component


try:
    class_to_test = os.environ["BMITEST_CLASS"]
except KeyError:
    Bmi = None
else:
    Bmi = load_component(class_to_test)

INPUT_FILE = os.environ.get("BMITEST_INPUT_FILE", None)


@pytest.mark.dependency()
def test_has_initialize():
    """Test component has an initialize method."""
    bmi = Bmi()
    assert hasattr(bmi, "initialize")


@pytest.mark.dependency()
def test_has_finalize():
    """Test component has a finalize method."""
    bmi = Bmi()
    assert hasattr(bmi, "finalize")


@pytest.mark.dependency(depends=["has_initialize", "has_finalize"], name="initialize_works")
def test_initialize(tmpdir):
    """Test component can initialize itself."""
    infile = os.environ.get("BMITEST_INPUT_FILE", None)
    manifest = os.environ.get("BMITEST_MANIFEST", infile or "").splitlines()

    with tmpdir.as_cwd() as prev:
        for file_ in [fname.strip() for fname in manifest]:
            if file_:
                cp(os.path.join(str(prev), file_), tmpdir / file_, create_dirs=True)

        bmi = Bmi()
        assert bmi.initialize(INPUT_FILE) is None
        bmi.finalize()


@pytest.mark.dependency(depends=["initialize_works"])
def test_update(tmpdir):
    """Test component can update itself."""
    infile = os.environ.get("BMITEST_INPUT_FILE", None)
    manifest = os.environ.get("BMITEST_MANIFEST", infile or "").splitlines()

    with tmpdir.as_cwd() as prev:
        for file_ in [fname.strip() for fname in manifest]:
            if file_:
                cp(os.path.join(str(prev), file_), tmpdir / file_, create_dirs=True)

        bmi = Bmi()
        bmi.initialize(INPUT_FILE)
        assert bmi.update() is None
        bmi.finalize()



import warnings

import pytest
import standard_names


def test_get_component_name(initialized_bmi):
    """Test component name is a string."""
    name = initialized_bmi.get_component_name()
    assert isinstance(name, str)

    return name


def test_var_names(var_name):
    """Test var names are valid."""
    assert isinstance(var_name, str)
    if standard_names.is_valid_name(var_name):
        standard_names.StandardName(var_name)
    else:
        warnings.warn("not a valid standard name: {name}".format(name=var_name))


@pytest.mark.dependency()
def test_input_var_name_count(initialized_bmi):
    if hasattr(initialized_bmi, "get_input_var_name_count"):
        n_names = initialized_bmi.get_input_var_name_count()
        assert isinstance(n_names, int)
        assert n_names >= 0
    else:
        pytest.skip("get_input_var_name_count not implemented")


@pytest.mark.dependency()
def test_output_var_name_count(initialized_bmi):
    if hasattr(initialized_bmi, "get_output_var_name_count"):
        n_names = initialized_bmi.get_output_var_name_count()
        assert isinstance(n_names, int)
        assert n_names >= 0
    else:
        pytest.skip("get_output_var_name_count not implemented")


@pytest.mark.dependency(depends=["test_input_var_name_count"])
def test_get_input_var_names(initialized_bmi):
    """Input var names is a list of strings."""
    names = initialized_bmi.get_input_var_names()
    assert isinstance(names, tuple)

    if hasattr(initialized_bmi, "get_input_var_name_count"):
        n_names = initialized_bmi.get_input_var_name_count()
        assert len(names) == n_names
    else:
        warnings.warn("get_input_var_name_count not implemented")


@pytest.mark.dependency(depends=["test_output_var_name_count"])
def test_get_output_var_names(initialized_bmi):
    """Output var names is a list of strings."""
    names = initialized_bmi.get_output_var_names()
    assert isinstance(names, tuple)

    if hasattr(initialized_bmi, "get_output_var_name_count"):
        n_names = initialized_bmi.get_output_var_name_count()
        assert len(names) == n_names
    else:
        warnings.warn("get_output_var_name_count not implemented")



import contextlib
import ctypes
import ctypes.util
import os
import sys

import numpy as np
import pkg_resources
import pytest
import six


def check_bmi(
    package,
    tests_dir=None,
    input_file=None,
    manifest=None,
    bmi_version="2.0",
    extra_args=None,
    help_pytest=False,
):
    if tests_dir is None:
        tests_dir = pkg_resources.resource_filename(__name__, "bootstrap")
    args = [tests_dir]

    os.environ["BMITEST_CLASS"] = package
    os.environ["BMITEST_INPUT_FILE"] = input_file or ""
    os.environ["BMI_VERSION_STRING"] = bmi_version

    if manifest:
        if isinstance(manifest, six.string_types):
            with open(manifest, "r") as fp:
                manifest = fp.read()
        else:
            manifest = os.linesep.join(manifest)
        os.environ["BMITEST_MANIFEST"] = manifest

    extra_args = extra_args or []
    if help_pytest:
        extra_args.append("--help")
    args += extra_args

    return pytest.main(args)


@contextlib.contextmanager
def suppress_stdout(streams):
    null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
    # Save the actual stdout (1) and stderr (2) file descriptors.
    save_fds = [os.dup(1), os.dup(2)]

    os.dup2(null_fds[0], 1)
    os.dup2(null_fds[1], 2)

    yield

    # Re-assign the real stdout/stderr back to (1) and (2)
    os.dup2(save_fds[0], 1)
    os.dup2(save_fds[1], 2)
    # Close the null files
    for fd in null_fds + save_fds:
        os.close(fd)


def empty_var_buffer(bmi, var_name):
    """Create an empty value buffer for a BMI variable.

    Examples
    --------
    >>> import numpy as np
    >>> from bmi_tester.api import empty_var_buffer

    >>> class Bmi:
    ...     def get_var_nbytes(self, name):
    ...         return 128
    ...     def get_var_type(self, name):
    ...         return "int"

    >>> buffer = empty_var_buffer(Bmi(), "var-name")
    >>> np.any(buffer != 0)
    True
    >>> buffer[:] = 0
    >>> np.all(buffer == 0)
    True
    """
    nbytes = bmi.get_var_nbytes(var_name)
    dtype = np.dtype(bmi.get_var_type(var_name))

    values = np.frombuffer(np.random.bytes(nbytes), dtype=dtype).copy()

    return values


def check_units(units):
    path_to_udunits_lib = ctypes.util.find_library("udunits2")
    udunits = ctypes.CDLL(path_to_udunits_lib)

    ut_read_xml = udunits.ut_read_xml
    ut_read_xml.argtypes = (ctypes.c_char_p, )
    ut_read_xml.restype = ctypes.c_void_p

    ut_parse = udunits.ut_parse
    ut_parse.argtypes = (ctypes.c_void_p, ctypes.c_char_p, ctypes.c_int)
    ut_parse.restype = ctypes.c_void_p

    with suppress_stdout(sys.stderr):
        ut_system = ut_read_xml(None)
    return ut_parse(ut_system, units.encode("utf-8"), 0) is not None
